@@ -4,23 +4,22 @@ use crate::memory::{DefaultMemoryCreator, Memory};
 use crate::table::Table;
 use crate::traphandlers::Trap;
 use crate::vmcontext::{
-    VMBuiltinFunctionsArray, VMCallerCheckedAnyfunc, VMContext, VMGlobalDefinition,
-    VMSharedSignatureIndex,
+    VMBuiltinFunctionsArray, VMCallerCheckedAnyfunc, VMGlobalDefinition, VMSharedSignatureIndex,
 };
-use crate::Store;
+use crate::ModuleMemFds;
+use crate::{CompiledModuleId, Store};
 use anyhow::Result;
 use std::alloc;
 use std::any::Any;
 use std::convert::TryFrom;
-use std::marker;
 use std::ptr::{self, NonNull};
 use std::slice;
 use std::sync::Arc;
 use thiserror::Error;
 use wasmtime_environ::{
-    DefinedFuncIndex, DefinedMemoryIndex, DefinedTableIndex, EntityRef, EntitySet, FunctionInfo,
-    GlobalInit, HostPtr, MemoryInitialization, MemoryInitializer, Module, ModuleType, PrimaryMap,
-    SignatureIndex, TableInitializer, TrapCode, VMOffsets, WasmType, WASM_PAGE_SIZE,
+    DefinedFuncIndex, DefinedMemoryIndex, DefinedTableIndex, EntityRef, FunctionInfo, GlobalInit,
+    InitMemory, MemoryInitialization, MemoryInitializer, Module, ModuleType, PrimaryMap,
+    SignatureIndex, TableInitializer, TrapCode, WasmType, WASM_PAGE_SIZE,
 };
 
 #[cfg(feature = "pooling-allocator")]
@@ -34,11 +33,17 @@ pub use self::pooling::{
 /// Represents a request for a new runtime instance.
 pub struct InstanceAllocationRequest<'a> {
     /// The module being instantiated.
-    pub module: Arc<Module>,
+    pub module: &'a Arc<Module>,
+
+    /// The unique ID of the module being allocated within this engine.
+    pub unique_id: Option<CompiledModuleId>,
 
     /// The base address of where JIT functions are located.
     pub image_base: usize,
 
+    /// If using MemFD-based memories, the backing MemFDs.
+    pub memfds: Option<&'a Arc<ModuleMemFds>>,
+
     /// Descriptors about each compiled function, such as the offset from
     /// `image_base`.
     pub functions: &'a PrimaryMap<DefinedFuncIndex, FunctionInfo>,
@@ -374,20 +379,60 @@ fn check_memory_init_bounds(
     Ok(())
 }
 
-fn initialize_memories(
-    instance: &mut Instance,
-    initializers: &[MemoryInitializer],
-) -> Result<(), InstantiationError> {
-    for init in initializers {
-        instance
-            .memory_init_segment(
-                init.memory_index,
-                init.data.clone(),
-                get_memory_init_start(init, instance)?,
-                0,
-                init.data.end - init.data.start,
-            )
-            .map_err(InstantiationError::Trap)?;
+fn initialize_memories(instance: &mut Instance, module: &Module) -> Result<(), InstantiationError> {
+    let memory_size_in_pages =
+        &|memory| (instance.get_memory(memory).current_length as u64) / u64::from(WASM_PAGE_SIZE);
+
+    // Loads the `global` value and returns it as a `u64`, but sign-extends
+    // 32-bit globals which can be used as the base for 32-bit memories.
+    let get_global_as_u64 = &|global| unsafe {
+        let def = if let Some(def_index) = instance.module.defined_global_index(global) {
+            instance.global(def_index)
+        } else {
+            &*instance.imported_global(global).from
+        };
+        if module.globals[global].wasm_ty == WasmType::I64 {
+            *def.as_u64()
+        } else {
+            u64::from(*def.as_u32())
+        }
+    };
+
+    // Delegates to the `init_memory` method which is sort of a duplicate of
+    // `instance.memory_init_segment` but is used at compile-time in other
+    // contexts so is shared here to have only one method of memory
+    // initialization.
+    //
+    // This call to `init_memory` notably implements all the bells and whistles
+    // so errors only happen if an out-of-bounds segment is found, in which case
+    // a trap is returned.
+    let ok = module.memory_initialization.init_memory(
+        InitMemory::Runtime {
+            memory_size_in_pages,
+            get_global_as_u64,
+        },
+        &mut |memory_index, offset, data| {
+            // If this initializer applies to a defined memory but that memory
+            // doesn't need initialization, due to something like uffd or memfd
+            // pre-initializing it via mmap magic, then this initializer can be
+            // skipped entirely.
+            if let Some(memory_index) = module.defined_memory_index(memory_index) {
+                if !instance.memories[memory_index].needs_init() {
+                    return true;
+                }
+            }
+            let memory = instance.get_memory(memory_index);
+            let dst_slice =
+                unsafe { slice::from_raw_parts_mut(memory.base, memory.current_length) };
+            let dst = &mut dst_slice[usize::try_from(offset).unwrap()..][..data.len()];
+            dst.copy_from_slice(instance.wasm_data(data.clone()));
+            true
+        },
+    );
+    if !ok {
+        return Err(InstantiationError::Trap(Trap::wasm(
+            TrapCode::HeapOutOfBounds,
+        )));
     }
 
     Ok(())
@@ -397,16 +442,11 @@ fn check_init_bounds(instance: &mut Instance, module: &Module) -> Result<(), Ins
     check_table_init_bounds(instance, module)?;
 
     match &instance.module.memory_initialization {
-        MemoryInitialization::Paged { out_of_bounds, .. } => {
-            if *out_of_bounds {
-                return Err(InstantiationError::Link(LinkError(
-                    "memory out of bounds: data segment does not fit".into(),
-                )));
-            }
-        }
         MemoryInitialization::Segmented(initializers) => {
             check_memory_init_bounds(instance, initializers)?;
         }
+        // Statically validated already to have everything in-bounds.
+        MemoryInitialization::Paged { .. } => {}
     }
 
     Ok(())
@@ -429,33 +469,7 @@ fn initialize_instance(
     initialize_tables(instance, module)?;
 
     // Initialize the memories
-    match &module.memory_initialization {
-        MemoryInitialization::Paged { map, out_of_bounds } => {
-            for (index, pages) in map {
-                let memory = instance.memory(index);
-                let slice =
-                    unsafe { slice::from_raw_parts_mut(memory.base, memory.current_length) };
-
-                for (page_index, page) in pages {
-                    debug_assert_eq!(page.end - page.start, WASM_PAGE_SIZE);
-                    let start = (*page_index * u64::from(WASM_PAGE_SIZE)) as usize;
-                    let end = start + WASM_PAGE_SIZE as usize;
-                    slice[start..end].copy_from_slice(instance.wasm_data(page.clone()));
-                }
-            }
-
-            // Check for out of bound access after initializing the pages to maintain
-            // the expected behavior of the bulk memory spec.
-            if *out_of_bounds {
-                return Err(InstantiationError::Trap(Trap::wasm(
-                    TrapCode::HeapOutOfBounds,
-                )));
-            }
-        }
-        MemoryInitialization::Segmented(initializers) => {
-            initialize_memories(instance, initializers)?;
-        }
-    }
+    initialize_memories(instance, &module)?;
 
     Ok(())
 }
@@ -465,6 +479,7 @@ unsafe fn initialize_vmcontext(instance: &mut Instance, req: InstanceAllocationR
 
     if let Some(store) = req.store.as_raw() {
         *instance.interrupts() = (*store).vminterrupts();
+        *instance.epoch_ptr() = (*store).epoch_ptr();
         *instance.externref_activations_table() = (*store).externref_activations_table().0;
         instance.set_store(store);
     }
@@ -482,10 +497,8 @@ unsafe fn initialize_vmcontext(instance: &mut Instance, req: InstanceAllocationR
     }
 
     // Initialize the built-in functions
-    ptr::write(
-        instance.vmctx_plus_offset(instance.offsets.vmctx_builtin_functions_begin()),
-        VMBuiltinFunctionsArray::initialized(),
-    );
+    *instance.vmctx_plus_offset(instance.offsets.vmctx_builtin_functions()) =
+        &VMBuiltinFunctionsArray::INIT;
 
     // Initialize the imports
     debug_assert_eq!(req.imports.functions.len(), module.num_imported_funcs);
@@ -652,6 +665,7 @@ impl OnDemandInstanceAllocator {
         &self,
         module: &Module,
         store: &mut StorePtr,
+        memfds: Option<&Arc<ModuleMemFds>>,
     ) -> Result<PrimaryMap<DefinedMemoryIndex, Memory>, InstantiationError> {
         let creator = self
             .mem_creator
@@ -660,13 +674,24 @@ impl OnDemandInstanceAllocator {
         let num_imports = module.num_imported_memories;
         let mut memories: PrimaryMap<DefinedMemoryIndex, _> =
             PrimaryMap::with_capacity(module.memory_plans.len() - num_imports);
-        for plan in &module.memory_plans.values().as_slice()[num_imports..] {
+        for (memory_idx, plan) in module.memory_plans.iter().skip(num_imports) {
+            // Create a MemFdSlot if there is an image for this memory.
+            let defined_memory_idx = module
+                .defined_memory_index(memory_idx)
+                .expect("Skipped imports, should never be None");
+            let memfd_image = memfds.and_then(|memfds| memfds.get_memory_image(defined_memory_idx));
+
             memories.push(
-                Memory::new_dynamic(plan, creator, unsafe {
-                    store
-                        .get()
-                        .expect("if module has memory plans, store is not empty")
-                })
+                Memory::new_dynamic(
+                    plan,
+                    creator,
+                    unsafe {
+                        store
+                            .get()
+                            .expect("if module has memory plans, store is not empty")
+                    },
+                    memfd_image,
+                )
                 .map_err(InstantiationError::Resource)?,
             );
         }
@@ -689,26 +714,20 @@ unsafe impl InstanceAllocator for OnDemandInstanceAllocator {
         &self,
         mut req: InstanceAllocationRequest,
     ) -> Result<InstanceHandle, InstantiationError> {
-        let memories = self.create_memories(&req.module, &mut req.store)?;
+        let memories = self.create_memories(&req.module, &mut req.store, req.memfds)?;
         let tables = Self::create_tables(&req.module, &mut req.store)?;
 
         let host_state = std::mem::replace(&mut req.host_state, Box::new(()));
 
         let mut handle = {
-            let instance = Instance {
-                module: req.module.clone(),
-                offsets: VMOffsets::new(HostPtr, &req.module),
+            let instance = Instance::create_raw(
+                &req.module,
+                req.unique_id,
+                &*req.wasm_data,
                 memories,
                 tables,
-                dropped_elements: EntitySet::with_capacity(req.module.passive_elements.len()),
-                dropped_data: EntitySet::with_capacity(req.module.passive_data_map.len()),
                 host_state,
-                wasm_data: &*req.wasm_data,
-                vmctx_initialized: false,
-                vmctx: VMContext {
-                    _marker: marker::PhantomPinned,
-                },
-            };
+            );
             let layout = instance.alloc_layout();
             let instance_ptr = alloc::alloc(layout) as *mut Instance;
             if instance_ptr.is_null() {
