@@ -263,6 +263,7 @@ unsafe fn initialize_wasm_page(
 ) -> Result<()> {
     // Check for paged initialization and copy the page if present in the initialization data
     if let MemoryInitialization::Paged { map, .. } = &instance.module.memory_initialization {
+        let memory_index = instance.module().memory_index(memory_index);
         let pages = &map[memory_index];
 
         let pos = pages.binary_search_by_key(&(page_index as u64), |k| k.0);
@@ -438,6 +439,7 @@ mod test {
         Imports, InstanceAllocationRequest, InstanceLimits, ModuleLimits,
         PoolingAllocationStrategy, Store, StorePtr, VMSharedSignatureIndex,
     };
+    use std::sync::atomic::AtomicU64;
     use std::sync::Arc;
     use wasmtime_environ::{Memory, MemoryPlan, MemoryStyle, Module, PrimaryMap, Tunables};
 
@@ -465,8 +467,13 @@ mod test {
             ..Tunables::default()
         };
 
-        let instances = InstancePool::new(&module_limits, &instance_limits, &tunables)
-            .expect("should allocate");
+        let instances = InstancePool::new(
+            PoolingAllocationStrategy::Random,
+            &module_limits,
+            &instance_limits,
+            &tunables,
+        )
+        .expect("should allocate");
 
         let locator = FaultLocator::new(&instances);
 
@@ -546,6 +553,12 @@ mod test {
                 fn out_of_gas(&mut self) -> Result<(), anyhow::Error> {
                     Ok(())
                 }
+                fn epoch_ptr(&self) -> *const AtomicU64 {
+                    std::ptr::null()
+                }
+                fn new_epoch(&mut self) -> Result<u64, anyhow::Error> {
+                    Ok(0)
+                }
             }
             struct MockModuleInfo;
             impl crate::ModuleInfoLookup for MockModuleInfo {
@@ -566,24 +579,23 @@ mod test {
             for _ in 0..instances.max_instances {
                 handles.push(
                     instances
-                        .allocate(
-                            PoolingAllocationStrategy::Random,
-                            InstanceAllocationRequest {
-                                module: module.clone(),
-                                image_base: 0,
-                                functions,
-                                imports: Imports {
-                                    functions: &[],
-                                    tables: &[],
-                                    memories: &[],
-                                    globals: &[],
-                                },
-                                shared_signatures: VMSharedSignatureIndex::default().into(),
-                                host_state: Box::new(()),
-                                store: StorePtr::new(&mut mock_store),
-                                wasm_data: &[],
+                        .allocate(InstanceAllocationRequest {
+                            module: &module,
+                            memfds: None,
+                            unique_id: None,
+                            image_base: 0,
+                            functions,
+                            imports: Imports {
+                                functions: &[],
+                                tables: &[],
+                                memories: &[],
+                                globals: &[],
                             },
-                        )
+                            shared_signatures: VMSharedSignatureIndex::default().into(),
+                            host_state: Box::new(()),
+                            store: StorePtr::new(&mut mock_store),
+                            wasm_data: &[],
+                        })
                         .expect("instance should allocate"),
                 );
             }
